// Array Structure class

extern ofstream file;

// Since the implementation particulars are not salient to this
// assignment, we use a simple fixed-size array, with copying
// on insertion.
// Ideally, we'd use something more elegant and much faster for 
// the implementation, like a rotating buffer that allows
// constant-time insertion at either end, by holding a "head" and
// a "tail" index inside the class.
// If the size of the buffer is a power of two, wraparounds can
// be made very efficient.
// However, we here offer the abovementioned simpler solution.

template <class DataType>
class Array_Structure: public ImplementationClass <DataType>
{
private:
  //////////////////////////////////////////////////////////////
  // State
  // 
	int max;
	DataType * array;


public:
  //////////////////////////////////////////////////////////////
  // Constructors
  // 
	Array_Structure() : max(0), array(0) {}
	Array_Structure(int maxSize) :
    max(maxSize), array(new DataType [maxSize]) {}
	Array_Structure(const Array_Structure &source) :
    ImplementationClass <DataType> (source)
      { copyFrom(source); }
	// destructor frees array space
	~Array_Structure() {
    delete [] array;
  }

	// = operator
	Array_Structure & operator=(const Array_Structure &source) {
    if (&source!=this) {
      if (max<source.max) {
        // Resize this array.
        delete [] array;
        copyFrom(source);
      } else {
        copyDataFrom(source);
      }
    } else { file << "Warning: self-assignment\n"; }
    return *this;
  }


  //////////////////////////////////////////////////////////////
  // Behaviour methods
  // 

	// includes searches for an element an returns true if it finds it
	virtual bool includes(const DataType & element) const {
    for (int i=0;i!=size;i++)
      if (array[i]==element)
        return true;
    return false;
  }
	// adds a new element in the array, at the end.
  virtual void addLast (const DataType & element) {
	  if (size < max) 
      { array[size++] = element; }
	  else 
      { file << "overflow ignoring " << element << endl; }
  }

	// adds a new element in the array, at the front.
	virtual void addFirst (const DataType & element) {
    if (size < max) 
      { insertAtFront(element); } 
    else
      { file << "overflow ignoring " << element << endl; }
  }

	// removeFirst 
	virtual DataType removeFirst(DequeueEmptyHandler func_ptr) {
    return size==0 ? func_ptr() : removeFromFront();
  }

	// removeLast 
	virtual DataType removeLast(DequeueEmptyHandler func_ptr) {
	  DataType answer;

	  if (!size) { answer = func_ptr(); } 
	  else { answer = array[(size--) - 1]; }
	  
	  return answer;
  }

	// print outputs all elements in the array in the order they are stored
	virtual void print() const {
    int i;

	  file << "ArrayStructure contents: <";
  
	  for (i = 0; i < size-1; i++)
  	  { file << array[i] << " "; }
	  file << array[i] << ">"<< endl;
  }

private:
  //////////////////////////////////////////////////////////////
  // Implementation methods
  // 
  void copyDataFrom(const Array_Structure &source) {
    for (int i=0;i!=max;i++)
      array[i]=source.array[i];
    size=source.size;
  }

  // Precondition:  array is not pointing at allocated memory.
  void copyFrom(const Array_Structure &source) {
    array=new DataType [source.max];
    max=source.max;
    copyDataFrom(source);
  }

  // Precondition:  array is not full.
  void insertAtFront(const DataType &element) {
    for (int i = size - 1; i >= 0; i--)
      { array[i+1] = array[i]; }

    array[0] = element; 
    size++;
  }

  // Precondition:  array is not empty.
  DataType removeFromFront() {
		DataType answer = array[0]; 
		for (int i = 1; i < size; i++)
      array[i - 1] = array[i];
    size--;
    return answer;
  }
};
