// List Structure class

extern ofstream file;

// Since the implementation particulars are not salient to this
// assignment, we use a simple singly-linked list.
// Ideally, we'd probably want to use a doubly-linked list, so
// that removing from the tail of the list can also be efficient.
// Caching a tail pointer is another useful optimization, but
// again, that wasn't the point.
// We here offer the abovementioned straightforward solution.

template <class DataType>
class List_Structure : public ImplementationClass <DataType>
{
private:
  //////////////////////////////////////////////////////////////
  // Node structure
  // 
  struct Node 
  {
	  DataType data;
	  Node *next;
	  Node (DataType new_data = DataType(), Node * n = NULL)
		  : data(new_data), next(n) { }		// a constructor
  };

private:
  //////////////////////////////////////////////////////////////
  // State
  // 
	Node *head;

public:
  // default constructor
	List_Structure() : head(0) {}
  // copy constructor
	List_Structure(const List_Structure &source) {copy(source);}
  // destructor
	~List_Structure() {clear();}

  // assignment operator
	const List_Structure &operator= (const List_Structure &source) {
	  if (this == &source) 
		  file << "warning self-assignment" << endl;
	  else { clear (); copy (source); }
	  return *this;
  }

	// includes searches for an element an returns true if it finds it
	virtual bool includes(const DataType & element) const {
    for (Node *currentNode = head;
         currentNode;
         currentNode = currentNode -> next)
      if (currentNode -> data == element)
        return true;
    return false;
  }

		// add a new member at the end
	virtual void addLast(const DataType & element) {
	  ++size;
    (head ?
      getLastNode()->next : head) = new Node (element);
  }

		// add a new member at the beginning
	virtual void addFirst(const DataType & element) {
	  ++size;
	  head = new Node(element, head);
  }

		// remove a node from the beginning
	virtual DataType removeFirst(DequeueEmptyHandler emptyHandlerFn) {
	  Node * nodeToDelete = head;
	  DataType answer;

	  if (!nodeToDelete) {
      answer = emptyHandlerFn( );
    } else {
		  answer = nodeToDelete -> data;
		  head = head -> next;
		  delete nodeToDelete;
		  --size;
	  }
	  return answer;
  }

		// remove a node from the end
	virtual DataType removeLast(DequeueEmptyHandler emptyHandlerFn) {
	  DataType answer;

	  if (!head) { answer = emptyHandlerFn (); }
	  else {
      Node **addressOfLastNode;
      if (!head -> next) {
        addressOfLastNode = &head;
	    } else {
        Node *secondLast = getSecondLastNode();
        addressOfLastNode = &secondLast -> next;
	    }
      answer = (*addressOfLastNode) -> data;
      delete *addressOfLastNode;
      *addressOfLastNode = 0;
		  --size;
    }

	  return answer;
  }

  // print all members 
	virtual void print() const {
	  file << "ListStructure contents: <";

    Node *currentNode = head;

    if (size>=2) {
  	  for (; currentNode -> next; currentNode = currentNode -> next)
    	  { file << currentNode -> data << "->"; }
    }
	  file << currentNode -> data << ">"<< endl;
  }


private:
  //////////////////////////////////////////////////////////////
  // Implementation methods
  // 

  // make this List into a copy of <source>
	void copy (const List_Structure &source) {
	  Node * currentNode, * sourceNode = source.head;

	  size = source.size;

	  if (sourceNode)
	  {
		  head = new Node(sourceNode -> data);
		  currentNode = head;
      // Advance pointers
		  sourceNode = sourceNode -> next;
	  }
	  while (sourceNode)
	  {
		  currentNode -> next = new Node(sourceNode -> data);
      // Advance pointers
		  currentNode = currentNode -> next;
		  sourceNode = sourceNode -> next;
	  }
  }

  // destroy the list
	void clear () {
	  Node * toDelete = head;
	  while (toDelete)
	  {
		  head = head -> next;
		  delete toDelete;
		  toDelete = head;
	  }
  }
  // Answer 0 if the list is empty.
  Node *getLastNode() const {
    Node *searchNode = head;
  	for (searchNode = head; searchNode->next; searchNode = searchNode->next);
    return searchNode;
  }
  // Precondition:  there is at least two nodes in the list.
  Node *getSecondLastNode() const {
		Node * prev = head, * current = head -> next;
		while (current -> next)
		{ 
			prev = current;
			current = current -> next;
		}
    return prev;
  }
};  // End class declaration/definition
